@@ -2,19 +2,19 @@
 #
 # Author: Michael Hudson-Doyle <michael.hudson@linaro.org>
 #
-# This file is part of Launch Control.
+# This file is part of Lava Dashboard.
 #
-# Launch Control is free software: you can redistribute it and/or modify
+# Lava Dashboard is free software: you can redistribute it and/or modify
 # it under the terms of the GNU Affero General Public License version 3
 # as published by the Free Software Foundation
 #
-# Launch Control is distributed in the hope that it will be useful,
+# Lava Dashboard is distributed in the hope that it will be useful,
 # but WITHOUT ANY WARRANTY; without even the implied warranty of
 # MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU Affero General Public License
-# along with Launch Control.  If not, see <http://www.gnu.org/licenses/>.
+# along with Lava Dashboard.  If not, see <http://www.gnu.org/licenses/>.
 
 import json
 
@@ -87,7 +87,13 @@ class UserFiltersView(FilterView):
 class PublicFiltersView(FilterView):
 
     def get_queryset(self):
-        return TestRunFilter.objects.filter(public=True)
+        filters = TestRunFilter.objects.filter(public=True)
+        non_accessible_filters = []
+        for filter in filters:
+            if not filter.is_accessible_by(self.request.user):
+                non_accessible_filters.append(filter.id)
+
+        return filters.exclude(id__in=non_accessible_filters)
 
 
 @BreadCrumb("Filters and Subscriptions", parent=index)
@@ -95,6 +101,7 @@ def filters_list(request):
     public_view = PublicFiltersView(request, model=TestRunFilter, table_class=PublicFiltersTable)
     prefix = "public_"
     public_filters_table = PublicFiltersTable(
+        request.user,
         public_view.get_table_data(prefix),
         prefix=prefix
     )
@@ -111,6 +118,7 @@ def filters_list(request):
         user_view = UserFiltersView(request, model=TestRunFilter, table_class=UserFiltersTable)
         prefix = "user_"
         user_filters_table = UserFiltersTable(
+            request.user,
             user_view.get_table_data(prefix),
             prefix=prefix
         )
@@ -172,6 +180,10 @@ def filter_detail(request, username, name):
     if not request.user.is_superuser:
         if not qfilter.public and qfilter.owner != request.user:
             raise PermissionDenied()
+
+    if not qfilter.is_accessible_by(request.user):
+        raise PermissionDenied()
+
     if not request.user.is_authenticated():
         subscription = None
     else:
@@ -295,6 +307,9 @@ def filter_edit(request, username, name):
         if request.user.username != username:
             raise PermissionDenied()
     filter = get_object_or_404(TestRunFilter, owner__username=username, name=name)
+    if not filter.is_accessible_by(request.user):
+        raise PermissionDenied()
+
     return filter_form(
         request,
         BreadCrumbTrail.leading_to(filter_edit, name=name, username=username),
@@ -309,6 +324,9 @@ def filter_copy(request, username, name):
         if not filter.public and filter.owner != request.user:
             raise PermissionDenied()
 
+    if not filter.is_accessible_by(request.user):
+        raise PermissionDenied()
+
     return filter_form(
         request,
         BreadCrumbTrail.leading_to(filter_copy, name=name, username=username),
@@ -322,6 +340,9 @@ def filter_delete(request, username, name):
         if request.user.username != username:
             raise PermissionDenied()
     filter = get_object_or_404(TestRunFilter, owner__username=username, name=name)
+    if not filter.is_accessible_by(request.user):
+        raise PermissionDenied()
+
     if request.method == "POST":
         if 'yes' in request.POST:
             filter.delete()
@@ -369,7 +390,7 @@ def get_test_cases_json(request):
 
     test_cases = TestCase.objects.filter(
         test__test_runs__bundle__bundle_stream__testrunfilter__id=request.GET['id'],
-        test__id=request.GET['test_id']).exclude(units__exact='').distinct('test_case_id').order_by('test_case_id')
+        test__id=request.GET['test_id']).distinct('test_case_id').order_by('test_case_id')
 
     data = serializers.serialize('json', test_cases)
     return HttpResponse(data, content_type='application/json')
